#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#ifdef TAGGED
#include "treiber_tagged.h"
#else
#ifdef HAZARD
#include "treiber_hazard.h"
#else
#include "treiber.h"
#endif
#endif

/*
On PL-59043,
* no ABA fails for 10000 80

 */
static int verbose = 0 ;

/* Context */
typedef struct {
  int n ;
  int sum ;
  int ntasks ;
  stack_t *stack ;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;

void *pusher(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;

  for (int k = 0 ; k < q->n ; k++) {
    boxed_int_t *v = alloc_boxed_int(k+1) ;
    if (verbose) fprintf(stderr,"%02i PUSH: %i<%p>\n",p->id,v->v,v) ;
    push(q->stack,v) ;
    if (verbose) fprintf(stderr,"%02i PUSHED: %i\n",p->id,k+1) ;
  }
#ifdef HAZARD
  flush_free_list() ;
#endif  
  return NULL ;
}

void ppint(void *p) {
  boxed_int_t *q = (boxed_int_t *)p ;
  fprintf(stderr,"%i",q->v) ;  
}

void *popper(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  do {
    boxed_int_t *n ;
    while ((n = pop(q->stack,p->id)) == NULL) {
      if (verbose > 1) fprintf(stderr,"%02i LOOP\n", p->id);
      delay(100) ;
      if (q->ntasks <= 0) return NULL;
    }
    int v = n->v ;
    free_boxed_int(n) ;
    if (verbose) { 
      fprintf(stderr,"%02i POP: %i\n",p->id,v) ;
#ifdef TAGGED
      if (verbose > 1) pp_stack(q->stack,ppint) ;
#endif
    }    
    (void)__sync_fetch_and_add(&q->sum,v) ;
  } while (__sync_add_and_fetch(&q->ntasks,-1) > 0) ;
#ifdef HAZARD
  flush_free_list() ;
#endif  
  return NULL ;
}

int seq(int n) {
  int sum = 0 ;
  for (int k = 0 ; k < n ; k++) sum += k+1 ;
  return sum ;
}

void run(int n, int npop) {
  if (verbose) fprintf(stderr,"n=%i, npop=%i\n",n,npop) ;
  int npush = 2*npop ;
  int nprocs = npush+npop ;
  common_t c ;
  c.n = n ; c.sum = 0 ;
  c.ntasks = npush*n ; c.stack = alloc_stack(nprocs) ;

  pthread_t th[nprocs] ; ctx_t a[nprocs] ;
  for (int k = 0 ; k < npop ; k++) {
    ctx_t *p = &a[k] ; p->id = k ; p->common = &c ;
    create_thread(&th[k],popper,p) ;
  }
  for(int k = npop ; k < nprocs ; k++) {
    ctx_t *p = &a[k] ; p->id = k ; p->common = &c ;
    create_thread(&th[k],pusher,p) ;
  }

  for (int k = 0 ; k < nprocs ; k++) {
    (void)join_thread(&th[k]) ;
  }

  int ok = npush*seq(n) ;
  printf("%i %i\n",c.sum, ok) ;

  free_stack(c.stack) ;
}

void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [sz] [nprocs]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int n = 10000 ;
  int npop = 4 ;

  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    n = atoi(argv[0]) ;
  } else if (argc > 1) {
    n = atoi(argv[0]) ;
    npop = atoi(argv[1]) ;
  }
  init_heap() ;
  run(n,npop) ;
  return 0 ;
}
