#include <stdlib.h>
#include <stdio.h>
#include "basic.h"


static int verbose = 0 ;
static int loop = 0 ;


/* Barrier */


#ifdef MY
#include "final_barrier.c"

#else
typedef pthread_barrier_t barrier_t ;

inline static barrier_t *alloc_my_barrier(unsigned int sz) {
  return alloc_barrier(sz) ;
}

inline static void free_my_barrier(barrier_t *p) {
  free_barrier(p) ;
}

inline static void wait_my_barrier(barrier_t *p) {
  (void)wait_barrier(p) ;
}
#endif


typedef struct {
  int size ;
  barrier_t *b ;
  volatile int *x,*y;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;


void *T1(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int size = q->size ;
  int *t = calloc_check(size, sizeof(t[0]));
  for (int k = size-1 ; k >= 0 ; k--) {
    wait_my_barrier(q->b) ;
    q->x[k] = 1 ;
    int r1 = q->y[k] ;
    t[k] = r1 ;
  }
  return t ;
}

void *T2(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int size = q->size ;
  int *t = calloc_check(size, sizeof(t[0]));
  for (int k = size-1 ; k >= 0 ; k--) {
    wait_my_barrier(q->b) ;
    q->y[k] = 1 ;
    __sync_synchronize() ;
    int r2 = q->x[k] ;
    t[k] = r2 ;
  }
  return t ;
}


void run(int size) {
  barrier_t *b = alloc_my_barrier(2) ;

  for ( ; ; ) {
    int *x = calloc(size,sizeof(x[0])) ;
    int *y = calloc(size,sizeof(y[0])) ;

    common_t c ;
    c.size = size ; c.b = b ; c.x = x ; c.y = y ;
    pthread_t th1,th2 ;
    ctx_t a1,a2 ;
    
    a1.id = 1 ; a1.common = &c ;
    create_thread(&th1,T1,&a1) ;
    a2.id = 2 ; a2.common = &c ;
    create_thread(&th2,T2,&a2) ;

    int *r1 = join_thread(&th1) ;
    int *r2 = join_thread(&th2) ;
    int count[2][2] ;
    count[0][0] = count[0][1] = count[1][0] = count[1][1] = 0 ;
    for (int k = 0 ; k < size ; k++) {
      count[r1[k]][r2[k]]++ ;
    }
    printf("**********\n") ;
    int nz = 0 ;
    for (int k1 = 0 ; k1 < 2 ; k1++)
      for (int k2 = 0 ; k2 < 2 ; k2++) {
        int c = count[k1][k2] ;
        if (c > 0) nz++ ;
        printf("%-5i> r1=%i r2=%i\n",c,k1,k2) ;
      }
    free(x) ; free(y) ; free(r1) ; free(r2) ;
    if (nz == 4 && !loop) {
      free_my_barrier(b) ;
      return ;
    }
  }
  free_my_barrier(b) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 10000 ;

  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else if (*arg == 'l') loop++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    size = atoi(argv[0]) ;
  }
  run(size) ;
  return 0 ;
}
